unit MainUnit;

{$mode objfpc}{$H+}

interface

uses
	Classes, SysUtils, Forms, Graphics, Grids, ExtCtrls, LazUtf8, Math, fgl,

	Windows, jwaWindows,

	ArrayList, modulesutil, errors, mystrings,

	agent_types;

type

	{ TMainForm }

	TMainForm = class(TForm)
		Grid: TStringGrid;
		Timer: TTimer;
		procedure SetGridCellValue(ACol, ARow: Integer; Value: String);
		procedure UpdateGridColumnWidth(ColumnIndex: Integer);
		procedure UpdateGridColumnsWidth();
		procedure UpdateGrid;
		procedure ConfigureGrid;
		procedure UpdateTypeData;
		procedure LoadTypes(AFileName: String);
		procedure GridHeaderClick(Sender: TObject; IsColumn: Boolean; Index: Integer);
		procedure GridCompareCells(Sender: TObject; ACol, ARow, BCol, BRow: Integer; var Result: integer);
		procedure OnTimerTick(Sender: TObject);
		procedure FormCreate(Sender: TObject);
		procedure FormDestroy(Sender: TObject);
	private
		{ private declarations }
	public
		{ public declarations }
	end;

var
	MainForm: TMainForm;
	LastSortedColumn: Integer = -1;
	PrevWndProc: Windows.WNDPROC;
	GameTickCount: Dword = 0;

	WarWindow: HWND;
	WarProcId: Dword;
	WarProc: Handle;
	pGameBase: Pointer;

implementation

{$R *.lfm}

function Read(Source: Pointer; Buffer: Pointer; Size: SizeUInt): Boolean;
var
	Written: SizeUInt;
begin
	Result:= ReadProcessMemory(WarProc, Source, Buffer, Size, @Written) and ({%H-}Written = Size);
end;

function SizeToStr(Size: Dword): String;
const
	Suffixes: array [0..4] of String = ('byte(s)', 'KiB','MiB','GiB','TiB');
var
	level: Integer = 0;
	amount: Dword;
begin
	amount:= Size;

	while (amount >= 1024) and (level < Length(Suffixes)-1) do begin
		amount:= amount div 1024;
		level += 1;
	end;

	if level > 0
		then Result:= Format('%.2f %s', [
				Size / Math.Power(1024, level),
				Suffixes[level]
			])
		else Result:= Format('%d %s', [Size, Suffixes[level]]);
end;

const
	COLS_COUNT = 4;
	NUMBER_COLUMN_ID = 0;
	NAME_COLUMN_ID   = 1;
	COUNT_COLUMN_ID  = 2;
	MEMORY_COLUMN_ID = 3;

{ TMainForm }

procedure TMainForm.SetGridCellValue(ACol, ARow: Integer; Value: String);
var
	TextWidth: Integer;
begin
	Grid.Cells[ACol, ARow]:= Value;
	TextWidth:= Grid.Canvas.TextWidth(Value);
	if TextWidth > Grid.ColWidths[ACol] then begin
		Grid.ColWidths[ACol]:= TextWidth;
	end;
end;

procedure TMainForm.UpdateGridColumnWidth(ColumnIndex: Integer);
var
	OneCharWidth, HighestWidth: Integer;
	i: Integer;
begin
	OneCharWidth:= Grid.Canvas.TextWidth('_');

	HighestWidth:= 0;
	for i:= 0 to Grid.RowCount-1 do begin
		HighestWidth:= Max(
				Grid.Canvas.TextWidth(Grid.Cells[ColumnIndex, i]),
				HighestWidth
			);
	end;

	Grid.ColWidths[MEMORY_COLUMN_ID]:= HighestWidth + OneCharWidth;
end;

procedure TMainForm.UpdateGridColumnsWidth();
var
	i: Integer;
begin
	for i:= 0 to COLS_COUNT - 1 do begin
		UpdateGridColumnWidth(i);
	end;
end;

procedure TMainForm.UpdateGrid;
var
	i: Integer;
	AType: TAgentType;
begin
	Grid.BeginUpdate;

	for i:= 1 to Grid.RowCount-1 do begin
		AType:= TAgentType(Grid.Objects[0,i]);
		SetGridCellValue(COUNT_COLUMN_ID, i, IntToStr(AType.InstanceCount));
		SetGridCellValue(MEMORY_COLUMN_ID, i, SizeToStr(AType.InstanceCount * AType.BytesPerInstance));
	end;

	if LastSortedColumn <> -1 then
		Grid.SortColRow(true, LastSortedColumn);

	Grid.EndUpdate;
end;

procedure TMainForm.ConfigureGrid;
var
	i: Integer;
begin
	Grid.RowCount:= Types.Count + 1;
	Grid.ColCount:= COLS_COUNT;

	for i:= 0 to Types.Count - 1 do begin
		Grid.Cells[NAME_COLUMN_ID,i+1]:= Types[i].Name;
		Grid.Objects[0,i+1]:= Types[i];
	end;

	UpdateGridColumnsWidth();
end;

procedure TMainForm.UpdateTypeData;
var
	i: Integer;
	Count: SizeUInt;
begin
	for i:= 0 to Types.Count - 1 do begin
		if Types[i] = nil then continue;
		if Types[i].AllocatorOffset = 0 then continue;
		Read(pGameBase + Types[i].AllocatorOffset + $0C, @Count, SizeOf(Count));
		Types[i].InstanceCount:= Count;
	end;
end;

procedure TMainForm.LoadTypes(AFileName: String);
var
	Strings: TStrings;
	NewType: TAgentType;
	Line: String;
	Parts: TStringsArray;
begin
	Strings:= TStringList.Create;
	try
		Strings.LoadFromFile(AFileName);

		Types.Capacity:= Strings.Count;
		Types.Count:= 0;

		for Line in Strings do begin
			Parts:= Split(Line, ' ', 2);
			NewType:= TAgentType.Create;
			NewType.Name:= Parts[0];
			NewType.AllocatorOffset:= StrToInt(Parts[1]);
			NewType.BytesPerInstance:= StrToInt(Parts[2]);
			Types.Add(NewType);
		end;
	except
		on e: Exception do begin
			e.Message:= Format(
				'Failed to load types data from «%s».'#13#10'%s',
				[
					AFileName,
					e.Message
				]
			);

			raise;
		end;
	end;

	Strings.Destroy;
end;

procedure TMainForm.GridHeaderClick(Sender: TObject; IsColumn: Boolean; Index: Integer);
begin
  if IsColumn then LastSortedColumn:= Index;
end;

generic function Compare<T>(const a, b: T): Integer;
begin
	if a < b
		then Exit(-1)
	else if a > b
		then Exit(1);

	Exit(0);
end;

procedure TMainForm.GridCompareCells(Sender: TObject; ACol, ARow, BCol, BRow: Integer; var Result: integer);
var
	AType, BType: TAgentType;
begin
	AType:= TAgentType(Grid.Objects[0,ARow]);
	BType:= TAgentType(Grid.Objects[0,BRow]);

	case ACol of
		NUMBER_COLUMN_ID: ;
		NAME_COLUMN_ID: Result:= UTF8CompareText(Grid.Cells[ACol,ARow], Grid.Cells[BCol,BRow]);
		COUNT_COLUMN_ID: Result:= AType.InstanceCount - BType.InstanceCount;
		MEMORY_COLUMN_ID: Result:= specialize Compare<Qword>(
				Qword(AType.InstanceCount) * AType.BytesPerInstance,
				Qword(BType.InstanceCount) * BType.BytesPerInstance
			);
	end;

	if Result = 0 then begin
		Result:= THandle(AType) - THandle(BType);
	end;

	if Grid.SortOrder = soDescending then
		Result:= -Result;
end;

procedure TMainForm.OnTimerTick(Sender: TObject);
begin
	UpdateTypeData;
	UpdateGrid;
end;

function EnableDebugPrivileges: Boolean;
var
	PrivilegeID: LUID;
	Token: Handle;
	Privileges: TTokenPrivileges;
begin
	Result:= False;

	if not LookupPrivilegeValue(nil, SE_DEBUG_NAME, {%H-}PrivilegeID) then exit;

	if not OpenProcessToken(
			GetCurrentProcess(),
			TOKEN_QUERY or TOKEN_ADJUST_PRIVILEGES,
			{%H-}Token
		)
		then exit;

	with Privileges do begin
		PrivilegeCount:= 1;

		with Privileges[0] do begin
			Luid:= PrivilegeID;
			Attributes:= SE_PRIVILEGE_ENABLED;
		end;
	end;

	if not AdjustTokenPrivileges(
			Token,
			False,
			@Privileges,
			SizeOf(Privileges),
			nil,
			nil
		)
		then exit;

	if GetLastError() <> ERROR_SUCCESS then exit;

    CloseHandle(Token);

	Result:= True;
end;

procedure TMainForm.FormCreate(Sender: TObject);
const
	WAR3_WINDOW_TITLE = 'Warcraft III';
var
	ModuleInfo: TModuleEntry32W;
begin
	EnableDebugPrivileges;

	WarWindow:= FindWindow(nil, WAR3_WINDOW_TITLE);
	if WarWindow = 0
		then RaiseLastOSError('Failed to find window with title «%s».', [WAR3_WINDOW_TITLE]);

	if GetWindowThreadProcessId(WarWindow, @WarProcId) = 0
		then RaiseLastOSError('Failed to get the game window process id.');

	WarProc:= OpenProcess(
		PROCESS_VM_READ,
		False,
		WarProcId
	);
	if WarProc = 0
		then RaiseLastOSError('Failed to open the game process.');

	ModuleInfo:= GetModule(WarProcId, 'Game.dll');
	pGameBase:= ModuleInfo.modBaseAddr;

	LoadTypes('.\types.txt');
	ConfigureGrid;
	UpdateGrid;
	Timer.Enabled:= True;;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
var
	AType: TAgentType;
begin
	for AType in Types do begin
		Types.Remove(AType);
		AType.Free;
	end;
end;

end.

















